#ifndef OALSOUND_H
#define OALSOUND_H

#ifndef NO_OPENAL

#include <vector>
#include <map>

#include "i_sound.h"
#include "s_sound.h"

#ifdef _WIN32
#include <al.h>
#include <alc.h>
#include <alure.h>
#elif defined(__APPLE__)
#include <OpenAL/al.h>
#include <OpenAL/alc.h>
#include <OpenAL/alure.h>
#else
#include <AL/al.h>
#include <AL/alc.h>
#include <AL/alure.h>
#endif

#include "efx.h"


class OpenALSoundStream;

class OpenALSoundRenderer : public SoundRenderer
{
public:
	OpenALSoundRenderer ();
	virtual ~OpenALSoundRenderer ();

	virtual void SetSfxVolume (float volume);
	virtual void SetMusicVolume (float volume);
	virtual SoundHandle LoadSound(BYTE *sfxdata, int length);
	virtual SoundHandle LoadSoundRaw(BYTE *sfxdata, int length, int frequency, int channels, int bits);
	virtual void UnloadSound (SoundHandle sfx);
	virtual unsigned int GetMSLength(SoundHandle sfx);
	virtual unsigned int GetSampleLength(SoundHandle sfx);
	virtual float GetOutputRate();

	// Streaming sounds.
	virtual SoundStream *CreateStream (SoundStreamCallback callback, int buffbytes, int flags, int samplerate, void *userdata);
	virtual SoundStream *OpenStream (const char *filename, int flags, int offset, int length);

	// Starts a sound.
	virtual FISoundChannel *StartSound (SoundHandle sfx, float vol, int pitch, int chanflags, FISoundChannel *reuse_chan);
	virtual FISoundChannel *StartSound3D (SoundHandle sfx, SoundListener *listener, float vol, FRolloffInfo *rolloff, float distscale, int pitch, int priority, const FVector3 &pos, const FVector3 &vel, int channum, int chanflags, FISoundChannel *reuse_chan);

	// Stops a sound channel.
	virtual void StopChannel (FISoundChannel *chan);

	// Returns position of sound on this channel, in samples.
	virtual unsigned int GetPosition(FISoundChannel *chan);

	// Synchronizes following sound startups.
	virtual void Sync (bool sync);

	// Pauses or resumes all sound effect channels.
	virtual void SetSfxPaused (bool paused, int slot);

	// Pauses or resumes *every* channel, including environmental reverb.
	virtual void SetInactive(bool inactive);

	// Updates the volume, separation, and pitch of a sound channel.
	virtual void UpdateSoundParams3D (SoundListener *listener, FISoundChannel *chan, bool areasound, const FVector3 &pos, const FVector3 &vel);

	virtual void UpdateListener (SoundListener *);
	virtual void UpdateSounds ();

	virtual bool IsValid ();
	virtual void PrintStatus ();
	virtual void PrintDriversList ();
	virtual FString GatherStats ();

private:
#define MAKE_FUNC(x) static typeof(::x)* x
	// Standard OpenAL 1.1 functions
	static void *openal_handle;
	MAKE_FUNC(alcCreateContext);
	MAKE_FUNC(alcMakeContextCurrent);
	MAKE_FUNC(alcProcessContext);
	MAKE_FUNC(alcSuspendContext);
	MAKE_FUNC(alcDestroyContext);
	MAKE_FUNC(alcGetCurrentContext);
	MAKE_FUNC(alcGetContextsDevice);
	MAKE_FUNC(alcOpenDevice);
	MAKE_FUNC(alcCloseDevice);
	MAKE_FUNC(alcGetError);
	MAKE_FUNC(alcIsExtensionPresent);
	MAKE_FUNC(alcGetProcAddress);
	MAKE_FUNC(alcGetEnumValue);
	MAKE_FUNC(alcGetString);
	MAKE_FUNC(alcGetIntegerv);
	MAKE_FUNC(alcCaptureOpenDevice);
	MAKE_FUNC(alcCaptureCloseDevice);
	MAKE_FUNC(alcCaptureStart);
	MAKE_FUNC(alcCaptureStop);
	MAKE_FUNC(alcCaptureSamples);
	MAKE_FUNC(alEnable);
	MAKE_FUNC(alDisable);
	MAKE_FUNC(alIsEnabled);
	MAKE_FUNC(alGetString);
	MAKE_FUNC(alGetBooleanv);
	MAKE_FUNC(alGetIntegerv);
	MAKE_FUNC(alGetFloatv);
	MAKE_FUNC(alGetDoublev);
	MAKE_FUNC(alGetBoolean);
	MAKE_FUNC(alGetInteger);
	MAKE_FUNC(alGetFloat);
	MAKE_FUNC(alGetDouble);
	MAKE_FUNC(alGetError);
	MAKE_FUNC(alIsExtensionPresent);
	MAKE_FUNC(alGetProcAddress);
	MAKE_FUNC(alGetEnumValue);
	MAKE_FUNC(alListenerf);
	MAKE_FUNC(alListener3f);
	MAKE_FUNC(alListenerfv);
	MAKE_FUNC(alListeneri);
	MAKE_FUNC(alGetListenerf);
	MAKE_FUNC(alGetListener3f);
	MAKE_FUNC(alGetListenerfv);
	MAKE_FUNC(alGetListeneri);
	MAKE_FUNC(alGetListeneriv);
	MAKE_FUNC(alGenSources);
	MAKE_FUNC(alDeleteSources);
	MAKE_FUNC(alIsSource);
	MAKE_FUNC(alSourcef);
	MAKE_FUNC(alSource3f);
	MAKE_FUNC(alSourcefv);
	MAKE_FUNC(alSourcei);
	MAKE_FUNC(alGetSourcef);
	MAKE_FUNC(alGetSource3f);
	MAKE_FUNC(alGetSourcefv);
	MAKE_FUNC(alGetSourcei);
	MAKE_FUNC(alGetSourceiv);
	MAKE_FUNC(alSourcePlayv);
	MAKE_FUNC(alSourceStopv);
	MAKE_FUNC(alSourceRewindv);
	MAKE_FUNC(alSourcePausev);
	MAKE_FUNC(alSourcePlay);
	MAKE_FUNC(alSourceStop);
	MAKE_FUNC(alSourceRewind);
	MAKE_FUNC(alSourcePause);
	MAKE_FUNC(alSourceQueueBuffers);
	MAKE_FUNC(alSourceUnqueueBuffers);
	MAKE_FUNC(alGenBuffers);
	MAKE_FUNC(alDeleteBuffers);
	MAKE_FUNC(alIsBuffer);
	MAKE_FUNC(alBufferData);
	MAKE_FUNC(alGetBufferf);
	MAKE_FUNC(alGetBufferfv);
	MAKE_FUNC(alGetBufferi);
	MAKE_FUNC(alGetBufferiv);
	MAKE_FUNC(alDopplerFactor);
	MAKE_FUNC(alDopplerVelocity);
	MAKE_FUNC(alDistanceModel);
	MAKE_FUNC(alListener3i);
	MAKE_FUNC(alListeneriv);
	MAKE_FUNC(alGetListener3i);
	MAKE_FUNC(alSource3i);
	MAKE_FUNC(alSourceiv);
	MAKE_FUNC(alGetSource3i);
	MAKE_FUNC(alBufferf);
	MAKE_FUNC(alBuffer3f);
	MAKE_FUNC(alBufferfv);
	MAKE_FUNC(alBufferi);
	MAKE_FUNC(alBuffer3i);
	MAKE_FUNC(alBufferiv);
	MAKE_FUNC(alGetBuffer3f);
	MAKE_FUNC(alGetBuffer3i);
	MAKE_FUNC(alSpeedOfSound);
	// ALURE 1.0 functions
	static void *alure_handle;
	MAKE_FUNC(alureGetErrorString);
	MAKE_FUNC(alureGetVersion);
	MAKE_FUNC(alureGetDeviceNames);
	MAKE_FUNC(alureFreeDeviceNames);
	MAKE_FUNC(alureInitDevice);
	MAKE_FUNC(alureShutdownDevice);
	MAKE_FUNC(alureGetSampleFormat);
	MAKE_FUNC(alureSleep);
	MAKE_FUNC(alureCreateBufferFromFile);
	MAKE_FUNC(alureCreateBufferFromMemory);
	MAKE_FUNC(alureBufferDataFromFile);
	MAKE_FUNC(alureBufferDataFromMemory);
	MAKE_FUNC(alureCreateStreamFromFile);
	MAKE_FUNC(alureCreateStreamFromMemory);
	MAKE_FUNC(alureCreateStreamFromStaticMemory);
	MAKE_FUNC(alureCreateStreamFromCallback);
	MAKE_FUNC(alureBufferDataFromStream);
	MAKE_FUNC(alureRewindStream);
	MAKE_FUNC(alureDestroyStream);
	MAKE_FUNC(alureInstallDecodeCallbacks);
#undef MAKE_FUNC

	// EFX Extension function pointer variables. Loaded after context creation
	// if EFX is supported. These pointers may be context-dependant, thus can't
	// be static
	// Effect objects
	LPALGENEFFECTS alGenEffects;
	LPALDELETEEFFECTS alDeleteEffects;
	LPALISEFFECT alIsEffect;
	LPALEFFECTI alEffecti;
	LPALEFFECTIV alEffectiv;
	LPALEFFECTF alEffectf;
	LPALEFFECTFV alEffectfv;
	LPALGETEFFECTI alGetEffecti;
	LPALGETEFFECTIV alGetEffectiv;
	LPALGETEFFECTF alGetEffectf;
	LPALGETEFFECTFV alGetEffectfv;
	// Filter objects
	LPALGENFILTERS alGenFilters;
	LPALDELETEFILTERS alDeleteFilters;
	LPALISFILTER alIsFilter;
	LPALFILTERI alFilteri;
	LPALFILTERIV alFilteriv;
	LPALFILTERF alFilterf;
	LPALFILTERFV alFilterfv;
	LPALGETFILTERI alGetFilteri;
	LPALGETFILTERIV alGetFilteriv;
	LPALGETFILTERF alGetFilterf;
	LPALGETFILTERFV alGetFilterfv;
	// Auxiliary slot objects
	LPALGENAUXILIARYEFFECTSLOTS alGenAuxiliaryEffectSlots;
	LPALDELETEAUXILIARYEFFECTSLOTS alDeleteAuxiliaryEffectSlots;
	LPALISAUXILIARYEFFECTSLOT alIsAuxiliaryEffectSlot;
	LPALAUXILIARYEFFECTSLOTI alAuxiliaryEffectSloti;
	LPALAUXILIARYEFFECTSLOTIV alAuxiliaryEffectSlotiv;
	LPALAUXILIARYEFFECTSLOTF alAuxiliaryEffectSlotf;
	LPALAUXILIARYEFFECTSLOTFV alAuxiliaryEffectSlotfv;
	LPALGETAUXILIARYEFFECTSLOTI alGetAuxiliaryEffectSloti;
	LPALGETAUXILIARYEFFECTSLOTIV alGetAuxiliaryEffectSlotiv;
	LPALGETAUXILIARYEFFECTSLOTF alGetAuxiliaryEffectSlotf;
	LPALGETAUXILIARYEFFECTSLOTFV alGetAuxiliaryEffectSlotfv;


	ALCcontext *Context;

	bool SrcDistanceModel;

	std::vector<ALuint> Sources;

	ALfloat SfxVolume;
	ALfloat MusicVolume;

	int SFXPaused;
	std::vector<ALuint> FreeSfx;
	std::vector<ALuint> PausableSfx;
	std::vector<ALuint> SfxGroup;

	const ReverbContainer *PrevEnvironment;

	typedef std::map<WORD,ALuint> EffectMap;
	ALuint EnvSlot;
	EffectMap EnvEffects;

	ALuint WaterLP;
	float LastWaterLP;

	std::vector<OpenALSoundStream*> Streams;
	friend class OpenALSoundStream;
};

#endif // NO_OPENAL

#endif
